using System;
using System.IO;
using System.Text;

namespace GIRSplit
{
	/// <summary>
	/// Splits Garmin MobilePC GIR files into their component PNGs
	/// </summary>
	class GIRSplitter
	{
        public static int iPngOffset = 0;
        public static int iPngSize = 0;
        public static int iPngWidth = 0;
        public static int iPngHeight = 0;

        [STAThread]
		static void Main(string[] args)
		{
			ParseInput(args);
		}

        //Check validity of command line input.
        public static void ParseInput(string[] args)
		{
            try
            {
                if (args.Length == 0)
                {
                    Console.WriteLine("");
                    Console.WriteLine("=============================================");
                    Console.WriteLine("GIRSplit: Splits and merges Garmin GIR files.");
                    Console.WriteLine("---------------------------------------------");
                    Console.WriteLine("  Usage: GIRSplit [filename.ext]\n");
                    Console.WriteLine("---------------------------------------------");
                    Console.WriteLine("Splits if the input file is \".gir\".");
                    Console.WriteLine("Merges if the input file is \".log\".");
                    Console.WriteLine("=============================================");
                    Console.WriteLine("");
                    Environment.Exit(1);
                }

                if (File.Exists(args[0]))
                {
                    if (args[0].Split('.')[1].ToLower() == "gir") { ParseSplit(args[0]); }
                    if (args[0].Split('.')[1].ToLower() == "log") { ParseMerge(args[0]); }
                }
                else
                {
                    Console.WriteLine("Error: File " + args[0] + " not found.\n");
                    Environment.Exit(1);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
		}

        //Splitter Functions
        public static void ParseSplit(string sInFile)
		{
			//Init Loc variables as -1 for use as error codes
			int iFirstPNG = -1;

            try
            {
                //Open the file for reading.
                FileStream fRead = new FileStream(sInFile, FileMode.Open, FileAccess.Read);

                //Create byte array to hold file contents.
                //Read the file into the array.
                byte[] bFileData = new byte[fRead.Length];
                int iResult = fRead.Read(bFileData, 0, Convert.ToInt32(fRead.Length));

                //Get the total image count.
                int iTotalImgs = BitConverter.ToInt32(bFileData, 0xC);

                //Define the signatures we are looking for.
                //Our damned while loop in IndexOfPattern() has a problem.
                //It executes one last time after the condition is false.
                //So I padded the signatures with 0xFF at the end to ensure a complete match.
                byte[] bPngHdr = { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0xFF };

                //Search the file for our pattern and return the results.
                iFirstPNG = IndexOfPattern(bFileData, bPngHdr, 0);
                if (iFirstPNG == -1)
                {
                    Console.WriteLine("PNG signature not found.."); //Nope, didn't find it.
                }
                else
                {
                    //We found it.. Now notify and look for the next one.
                    Console.WriteLine("First PNG signature found at: 0x" + String.Format("{0:x2}", iFirstPNG));
                }

                //SplitFile(fRead.Name, bFileData, iFirstPNG, iCramfsLoc);

                //Get the names block into an array.
                int iNamesSize = iFirstPNG - 0x10;
                byte[] bNamesBlock = new byte[iNamesSize];
                fRead.Position = 0x10;
                iResult = fRead.Read(bNamesBlock, 0, iNamesSize);

                //Get the PNG files block into an array.
                int iFilesSize = Convert.ToInt32(fRead.Length) - iNamesSize - 0x10;
                bFileData = new byte[iFilesSize];
                fRead.Position = iFirstPNG;
                iResult = fRead.Read(bFileData, 0, iFilesSize);

                //Create a subdirectory to hold the PNG files
                FileInfo fInfo = new FileInfo(fRead.Name);
                string sPath = fInfo.DirectoryName;
                string sBaseName = Path.GetFileNameWithoutExtension(fInfo.FullName);
                string sDumpDirectory = sPath + "\\" + sBaseName + "\\";
                Directory.CreateDirectory(sDumpDirectory);

                //Cycle through all the PNGs.
                for (int i = 0; i < iTotalImgs; i++)
                {
                    //Offset to a record in the names block array.
                    int iNameRecOffset = i * 0x38;

                    //Find the end of the name in the record.
                    //Names are zero terminated and a maximum of 0x28 chars.
                    int iNameEnd = Array.IndexOf(bNamesBlock, (byte)0, iNameRecOffset) - iNameRecOffset;
                    if (iNameEnd > 0x28) { iNameEnd = 0x28; }

                    //Get the name, img-dimension/position-data, unknown-word, array-offset-to-image and size of PNG.
                    string outFileName = Encoding.ASCII.GetString(bNamesBlock, iNameRecOffset, iNameEnd);
                    string sOutLog = //Offset word
                                     String.Format("{0:x2}", bNamesBlock[iNameRecOffset + 0x2C]).ToUpper() +
                                     String.Format("{0:x2}", bNamesBlock[iNameRecOffset + 0x2D]).ToUpper() +
                                     String.Format("{0:x2}", bNamesBlock[iNameRecOffset + 0x2E]).ToUpper() +
                                     String.Format("{0:x2}", bNamesBlock[iNameRecOffset + 0x2F]).ToUpper() +
                                     "|" + outFileName;
                    int iImgOffset = BitConverter.ToInt32(bNamesBlock, iNameRecOffset + 0x30);
                    int iPngSize = BitConverter.ToInt32(bNamesBlock, iNameRecOffset + 0x34);

                    //Create and fill a byte array with the PNG image.
                    byte[] bPngOutFile = new byte[iPngSize];
                    Array.Copy(bFileData, iImgOffset, bPngOutFile, 0, iPngSize);

                    //Dump the PNG data to a file.
                    SplitFile(sBaseName, outFileName, sDumpDirectory, sOutLog, bPngOutFile);
                }

                //Close the file handle.
                fRead.Close();

                //We're all done here..
                Console.WriteLine("All done...");
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }
		public static void SplitFile(string baseName, string outFileName, string dumpDir, string sOuputLog, byte[] outFileBytes)
		{
            try
            {
                //Create PNG File
                Console.WriteLine("Splitting out " + outFileName);
                FileStream hdrWrite = new FileStream(dumpDir + outFileName + ".PNG", FileMode.Create);
                hdrWrite.Write(outFileBytes, 0, outFileBytes.Length);
                hdrWrite.Close();

                //Append to logfile
                FileStream logWrite = new FileStream(dumpDir + baseName + ".log", FileMode.Append);
                StreamWriter txtWrite = new StreamWriter(logWrite);
                txtWrite.WriteLine(sOuputLog);
                txtWrite.Close();
                logWrite.Close();

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
		}
		
        //Merge Functions
		public static void ParseMerge(string sInFile)
		{
            try
            {
                //Get the directory and filename for the GIR file to be created.
                FileInfo logInfo = new FileInfo(sInFile);
                string girDir = logInfo.DirectoryName + "\\";
                string girName = Path.GetFileNameWithoutExtension(logInfo.FullName);

                //Create the gir file for writing
                FileStream girWrite = new FileStream(girDir + girName + ".gir", FileMode.Create);

                //Open the file for reading.
                FileStream logRead = new FileStream(sInFile, FileMode.Open, FileAccess.Read);
                StreamReader txtRead = new StreamReader(logRead);

                //Init while loop vars.
                int iLineCount = 0;
                string str = "x";
                while (str != null)
                {
                    //Loop through the file to get a count of
                    //the files to be merged.
                    str = txtRead.ReadLine();
                    iLineCount++;
                }
                //There's a null line at the end of the file.
                iLineCount = iLineCount - 1;

                //Create MemoryStreams and a BinaryWriter.
                // The hdrBlock memstream will hold the file header info.
                // The pngBlock memstream will hold the PNG files.
                // The BinaryWriter will write bytes to the hdrBlock.
                MemoryStream hdrBlock = new MemoryStream();
                MemoryStream pngBlock = new MemoryStream();
                BinaryWriter hdrWriter = new BinaryWriter(hdrBlock);

                //Dump the static stuff to the header block.
                hdrWriter.Write(Encoding.Default.GetBytes("gir"));
                byte[] gHdr = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00 };
                hdrWriter.Write(gHdr);

                //Dump the first dynamic data (png file count) to the header block.
                hdrWriter.Write(BitConverter.GetBytes(iLineCount));

                //Reset the logfile start position so we can search through it again.
                logRead.Position = 0;
                string sLine = "x";
                while (sLine != null)
                {
                    //Read a line from the logfile.
                    sLine = txtRead.ReadLine();
                    if (sLine == null) { break; }

                    //Split the line into the filename and the unknown dword.
                    string sPngName = sLine.Split('|')[1];
                    string sPngOffset = sLine.Split('|')[0];

                    //Spit out some status.
                    Console.WriteLine("Merging " + sPngName);

                    //Pad the filename out to the required bytes if needed.
                    byte[] bPngName = PadBytes(sPngName, 0x28);

                    //Get the unknown dword into a byte array.
                    byte[] bUnkData = new byte[sPngOffset.Length / 2];
                    for (int x = 0; x < bUnkData.Length; x++) { bUnkData[x] = Convert.ToByte(sPngOffset.Substring(x * 2, 2), 16); }

                    //Call the parser to get the size into a byte array.
                    //The PngParser takes care of other functions as well.
                    byte[] bPngSize = PngParser(girDir, sPngName, ref pngBlock);

                    //Get the png offset into a byte array
                    byte[] bPngOffset = BitConverter.GetBytes(iPngOffset);

                    //Get the PNG width-height into a byte array.
                    byte[] bWidth = BitConverter.GetBytes(Convert.ToInt16(iPngWidth));
                    byte[] bHeight = BitConverter.GetBytes(Convert.ToInt16(iPngHeight));

                    //Dump the values into the hdrBlock memstream in the correct order.
                    hdrWriter.Write(bPngName);
                    hdrWriter.Write(bWidth);
                    hdrWriter.Write(bHeight);
                    hdrWriter.Write(bUnkData);
                    hdrWriter.Write(bPngOffset);
                    hdrWriter.Write(bPngSize);

                    //Adjust the png offset for the next pass.
                    iPngOffset = iPngOffset + iPngSize;
                }

                //Write the header and png data to the GIR file.
                hdrBlock.WriteTo(girWrite);
                pngBlock.WriteTo(girWrite);

                //Close the file handles.
                txtRead.Close();
                logRead.Close();
                hdrWriter.Close();
                hdrBlock.Close();
                pngBlock.Close();

                Console.WriteLine(iLineCount.ToString());

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }
        private static byte[] PadBytes(string bArry, int iLen)
        {
            byte[] bReturn = new byte[iLen];

            try
            {
                for (int i = 0; i < bArry.Length; i++)
                {
                    bReturn[i] = (byte)bArry[i];
                }
                return bReturn;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
                return bReturn;
            }
        }
        private static byte[] PngParser(string sDirName, string sFileName, ref MemoryStream pngBlock)
        {
            //Define filestream to read PNG file.
            FileStream pngFile = new FileStream(sDirName + sFileName + ".PNG", FileMode.Open, FileAccess.Read);
            
            //Define MemoryStream to hold PNG file contents.
            MemoryStream mStr = new MemoryStream();
            mStr.SetLength(pngFile.Length);

            //Read file into memory stream.
            pngFile.Read(mStr.GetBuffer(), 0, (int)pngFile.Length);

            //Write file into main PNG block.
            mStr.WriteTo(pngBlock);

            //Get image size for png offset.
            iPngSize = (int)pngFile.Length;

            //Convert the image filesize into bytes
            byte[] bRetSize = BitConverter.GetBytes(iPngSize);

            // Now we get to find the width and height of the png file.
            //Locate the PNG IHDR chunk.
            byte[] pngIHDR = { 0x49, 0x48, 0x44, 0x52 }; //ASCII "IHDR"
            int iHdrLoc = IndexOfPattern(mStr.GetBuffer(), pngIHDR, 0);

            //Get the PNG width from the first DWORD after IHDR.
            byte[] bWidth = new byte[4];
            Array.Copy(mStr.GetBuffer(), iHdrLoc + 4, bWidth, 0, 4);
            Array.Reverse(bWidth);
            iPngWidth = BitConverter.ToInt32(bWidth, 0);

            //Get the PNG height from the second DWORD after IHDR.
            byte[] bHeight = new byte[4];
            Array.Copy(mStr.GetBuffer(), iHdrLoc + 8, bHeight, 0, 4);
            Array.Reverse(bHeight);
            iPngHeight = BitConverter.ToInt32(bHeight, 0);
            
            //Close the streams and return the filesize byte array
            mStr.Close();
            pngFile.Close();
            return bRetSize;
        }

        //Common Function
        public static int IndexOfPattern(byte[] byteBlock, byte[] bytePattern, int startPos)
        {
            int _blockPos = startPos;
            int _patternPos = 0;
            int _blockLen = byteBlock.Length;
            int _patternLen = bytePattern.Length;
            int _foundPos = -1;
            int _endPos = _blockLen - _patternLen;

            int test = 0;
            int mtst = 0;

            // While spot in block less than length of block minus length of pattern, 
            // stop at first pattern match found [you can turn this around to do some- 
            // thing equivalent to Array::LastIndexOf( ).] 
            //
            try
            {
                while (_blockPos <= _endPos && _foundPos == -1)
                {
                    // May need to return to this position plus one; Found Position is the 
                    // index of the first matching byte of the Pattern within the Block. 
                    _foundPos = _blockPos;

                    // If pattern Position equals pattern Length then a pattern match is 
                    // complete.  Compare byte-by-byte (or object-by-object). 
                    while ((byteBlock[_blockPos++] == bytePattern[_patternPos++]) && (_patternPos < _patternLen))
                    {
                        test = byteBlock[_blockPos];
                        mtst = bytePattern[_patternPos];
                    }

                    if (_patternPos != _patternLen)     // If Not Found 
                    {
                        _patternPos = 0;  // Start checking pattern from beginning again. 
                        _blockPos = _foundPos + 1;    // Restore block position. 
                        _foundPos = -1;    // Tell outer while loop to keep searching. 
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

            return _foundPos;
        } 
    }
}
